public class RBTree {
    private static final NodeRB EMPTY = new NodeRB(0);
    static {
        EMPTY.leftNodeRB = EMPTY;
        EMPTY.rightNodeRB = EMPTY;
    }
    protected NodeRB current;
    private NodeRB parent;
    private NodeRB grand;
    private NodeRB ancestor;
    private final NodeRB root;

    static final byte RED = 0;
    static final byte BLACK = 1;


    public NodeRB getRoot() {
        return root;
    }

    public RBTree(int rootKey) {
        this.root = new NodeRB(rootKey);
        this.root.setLeftNodeRB(EMPTY);
        this.root.setRightNodeRB(EMPTY);
    }

    public void insert(int item) {
        current = parent = grand = root;
        EMPTY.setKey(item);
        while (current.getKey() != item) {
            ancestor = grand;
            grand = parent;
            parent = current;

            current = item > current.getKey() ? current.getRightNodeRB() : current.getLeftNodeRB();
            if (current.getLeftNodeRB().getColor() == RED && current.getRightNodeRB().getColor() == RED) {
                rebalance(item);
            }
        }
        if (current != EMPTY) {
            return;
        }
        current = new NodeRB(item, EMPTY, EMPTY);

        if (item < parent.getKey()) {
            parent.setLeftNodeRB(current);
        } else {
            parent.setRightNodeRB(current);
        }
        rebalance(item);
    }

    private void rebalance(int item) {
        current.setColor(RED);
        current.getLeftNodeRB().setColor(BLACK);
        current.getRightNodeRB().setColor(BLACK);

        if (parent.getColor() == RED) {
            grand.setColor(RED);
            if (item < grand.getKey() != item < parent.getKey()) {
                parent = rotate(item, grand);
            }
            current = rotate(item, ancestor);
            current.setColor(BLACK);
        }
        root.getRightNodeRB().setColor(BLACK);
    }

    private NodeRB rotate(int item, NodeRB parent) {
        if (item < parent.getKey()) {
            NodeRB node = parent.getLeftNodeRB();
            NodeRB result = item < node.getKey() ? rotateLeft(node) : rotateRight(node);
            parent.setLeftNodeRB(result);
            return parent.getLeftNodeRB();
        } else {
            NodeRB node = parent.getRightNodeRB();
            NodeRB result = item < node.getKey() ? rotateLeft(node) : rotateRight(node);
            parent.setRightNodeRB(result);
            return parent.getRightNodeRB();
        }
    }

    private NodeRB rotateLeft(NodeRB node) {
        var left = node.getLeftNodeRB();
        node.setLeftNodeRB(left.getRightNodeRB());
        left.setRightNodeRB(node);
        return left;
    }

    private NodeRB rotateRight(NodeRB node) {
        var right = node.getRightNodeRB();
        node.setRightNodeRB(right.getLeftNodeRB());
        right.setLeftNodeRB(node);
        return right;
    }
}