public class RedBlackTree {
    static final COLOR RED = COLOR.RED;
    static final COLOR BLACK = COLOR.BLACK;
    protected Node root;

    public Node getRoot() {
        return this.root;
    }

    public void insert(int key) {
        Node node = root;
        Node parent = null;

        // Traverse the tree to the left or right depending on the key
        while (node != null) {
            parent = node;
            if (key < node.getData()) {
                node = node.getLeft();
            } else if (key > node.getData()) {
                node = node.getRight();
            } else {
                throw new IllegalArgumentException("Duplicated key " + key);
            }
        }

        Node newNode = new Node(key);
        newNode.setColor(RED);
        if (parent == null) {
            root = newNode;
        } else if (key < parent.getData()) {
            parent.setLeft(newNode);
        } else {
            parent.setRight(newNode);
        }
        newNode.setParent(parent);
        recolor(newNode);
    }

    private void recolor(Node node) {
        Node parent = node.getParent();
        if (parent == null || parent.getColor() == BLACK) {
            return;
        }

        Node grandparent = parent.getParent();
        if (grandparent == null) {
            parent.setColor(BLACK);
            return;
        }

        Node uncle = getUncle(parent);
        if (uncle != null && uncle.getColor() == RED) {
            parent.setColor(BLACK);
            grandparent.setColor(RED);
            uncle.setColor(BLACK);
            recolor(grandparent);
        }

        else if (parent == grandparent.getLeft()) {
            if (node == parent.getRight()) {
                rotateLeft(parent);
                parent = node;
            }
            rotateRight(grandparent);
            parent.setColor(BLACK);
            grandparent.setColor(RED);
        }

        else {
            if (node == parent.getLeft()) {
                rotateRight(parent);
                parent = node;
            }
            rotateLeft(grandparent);
            parent.setColor(BLACK);
            grandparent.setColor(RED);
        }
    }

    private Node getUncle(Node parent) {
        Node grandparent = parent.getParent();
        return grandparent.getLeft() == parent ? grandparent.getRight() : grandparent.getLeft();
    }

    private void rotateRight(Node node) {
        Node parent = node.getParent();
        Node leftChild = node.getLeft();
        node.setLeft(leftChild.getRight());
        if (leftChild.getRight() != null) {
            leftChild.getRight().setParent(node);
        }
        leftChild.setRight(node);
        node.setParent(leftChild);
        replaceParentsChild(parent, node, leftChild);
    }

    private void rotateLeft(Node node) {
        Node parent = node.getParent();
        Node rightChild = node.getRight();
        node.setRight(rightChild.getLeft());
        if (rightChild.getLeft() != null) {
            rightChild.getLeft().setParent(node);
        }
        rightChild.setLeft(node);
        node.setParent(rightChild);
        replaceParentsChild(parent, node, rightChild);
    }
    private void replaceParentsChild(Node parent, Node oldChild, Node newChild) {
        if (parent == null) {
            root = newChild;
        } else if (parent.getLeft() == oldChild) {
            parent.setLeft(newChild);
        } else if (parent.getRight() == oldChild) {
            parent.setRight(newChild);
        } else {
            throw new IllegalStateException("Узел не является потомком");
        }
        if (newChild != null) {
            newChild.setParent(parent);
        }
    }

}
